import java.util.concurrent.BlockingQueue
import java.util.concurrent.LinkedBlockingQueue
import scala.collection.mutable.HashMap

/******************************************************************************/
/*                                CHANNEL STUFF                               */
/******************************************************************************/

trait Channel{
  def send(t : String) : Unit
  def receive() : String 
}

object StdoutChannel extends Channel {
  def send(t:String) {
    this.synchronized{
      println(t)
    }
  }
  def receive() : String = {
    scala.sys.error("trying to receive on stdout")
  }
}

class AsyncChannel(queue : BlockingQueue[String]) extends Channel {
  def this() =  this(new LinkedBlockingQueue[String])
  def send(t : String) {
    queue.put(t)
  }
  def receive() : String = {
    queue.take()
  }
}

class SyncChannel extends Channel {
  var message : Option[String] = None
  var receptionPending = false
  var receptionAcknowledged = true
  private def messagePending() : Boolean = {
    this.message match {
      case None => false
      case Some(_) => true
    }
  }
  private def getMessage() : String = {
    message match {
      case None => scala.sys.error("trying to get a non existent message")
      case Some(t) => {
	message = None;
	return t
      }
    }
  }
  def send(t : String){
    this.synchronized{
      while (this.messagePending() || !(receptionAcknowledged)){
	wait()
      };
      receptionAcknowledged = false;
      message = Some(t);
      notifyAll();
      while (this.messagePending()){
	wait()
      }
      receptionAcknowledged = true;
      notifyAll();
    }
  }
  def receive() : String = {
    this.synchronized{
      while (!(this.messagePending())) {
	wait()
      }
      val res = this.getMessage();
      notifyAll();
      return res
    }
  }
}

object ChannelManager {
  var synchronous = true
  val channelMap = HashMap[String, Channel]("stdout" -> StdoutChannel)
  def getChannel(id : String) : Channel = {
    val optionChan = channelMap.get(id)
    optionChan match{
      case Some(chan) => chan
      case None => {
	if (synchronous) {
	  val chan = new SyncChannel()
	  channelMap.put(id,chan);
	  chan
	}
	else{
	  val chan = new AsyncChannel()
	  channelMap.put(id,chan);
	  chan
	}
      }
    }
  }
}
